# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

from trytond.tests.test_tryton import ModuleTestCase


class SaleInvoiceGroupingTestCase(ModuleTestCase):
    'Test Sale Invoice Grouping module'
    module = 'sale_invoice_grouping'


del ModuleTestCase



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

from trytond.pool import Pool

from . import configuration, party, sale


def register():
    Pool.register(
        sale.Sale,
        party.Party,
        party.PartySaleInvoiceGroupingMethod,
        configuration.Configuration,
        configuration.ConfigurationSaleMethod,
        module='sale_invoice_grouping', type_='model')



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

from itertools import groupby

from trytond.pool import PoolMeta
from trytond.transaction import Transaction


class Sale(metaclass=PoolMeta):
    __name__ = 'sale.sale'

    @property
    def invoice_grouping_method(self):
        party = self.invoice_party or self.party
        return party.sale_invoice_grouping_method

    def _get_invoice_grouping_fields(self, invoice):
        return ['state', 'company', 'type', 'journal', 'party',
            'invoice_address', 'currency', 'account', 'payment_term']

    def _get_grouped_invoice_order(self):
        "Returns the order clause used to find invoice that should be grouped"
        return None

    def _get_grouped_invoice_domain(self, invoice):
        "Returns a domain that will find invoices that should be grouped"
        Invoice = invoice.__class__
        invoice_domain = [
            ('lines.origin', 'like', 'sale.line,%'),
            ]
        fields = self._get_invoice_grouping_fields(invoice)
        defaults = Invoice.default_get(fields, with_rec_name=False)
        for field in fields:
            invoice_domain.append(
                (field, '=', getattr(invoice, field, defaults.get(field)))
                )
        return invoice_domain

    def _get_invoice_sale(self):
        transaction = Transaction()
        context = transaction.context
        invoice = super(Sale, self)._get_invoice_sale()
        if (not context.get('skip_grouping', False)
                and self.invoice_grouping_method):
            with transaction.set_context(skip_grouping=True):
                invoice = self._get_invoice_sale()
            Invoice = invoice.__class__
            domain = self._get_grouped_invoice_domain(invoice)
            order = self._get_grouped_invoice_order()
            grouped_invoices = Invoice.search(domain, order=order, limit=1)
            if grouped_invoices:
                invoice, = grouped_invoices
        return invoice

    @classmethod
    def _process_invoice(cls, sales):
        for method, sales in groupby(
                sales, lambda s: s.invoice_grouping_method):
            if method:
                for sale in sales:
                    super()._process_invoice([sale])
            else:
                super()._process_invoice(list(sales))



# This file is part of Tryton.  The COPYRIGHT file at the top level of this
# repository contains the full copyright notices and license terms.
from trytond.model import ModelSQL, ValueMixin, fields
from trytond.pool import Pool, PoolMeta


class Party(metaclass=PoolMeta):
    __name__ = 'party.party'

    sale_invoice_grouping_method = fields.MultiValue(fields.Selection([
                (None, 'None'),
                ('standard', 'Standard'),
                ],
            'Sale Invoice Grouping Method'))
    sale_invoice_grouping_methods = fields.One2Many(
        'party.party.sale_invoice_grouping_method', 'party',
        "Sale Invoice Grouping Methods")

    @classmethod
    def default_sale_invoice_grouping_method(cls, **pattern):
        pool = Pool()
        Configuration = pool.get('sale.configuration')
        return Configuration(1).get_multivalue(
            'sale_invoice_grouping_method', **pattern)


class PartySaleInvoiceGroupingMethod(ModelSQL, ValueMixin):
    "Party Sale Invoice Grouping Method"
    __name__ = 'party.party.sale_invoice_grouping_method'
    party = fields.Many2One(
        'party.party', "Party", ondelete='CASCADE')
    sale_invoice_grouping_method = fields.Selection(
        'get_sale_invoice_grouping_methods', "Sale Invoice Grouping Method")

    @classmethod
    def get_sale_invoice_grouping_methods(cls):
        pool = Pool()
        Party = pool.get('party.party')
        field_name = 'sale_invoice_grouping_method'
        return Party.fields_get([field_name])[field_name]['selection']
